package com.example.clientauth.clientauth;

import android.os.Bundle;
import android.os.StrictMode;
import android.support.v7.app.AppCompatActivity;
import android.util.Base64;
import android.view.View;
import android.widget.TextView;

import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.StringWriter;
import java.net.URL;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.Certificate;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLHandshakeException;
import javax.net.ssl.SSLSession;
import javax.net.ssl.TrustManagerFactory;

public class ClientAuth extends AppCompatActivity {
    private Integer requestNumber = 0;
    private SSLContext context;
    private KeyStore generatedKeyStore;
    private char[] password = "uselesspw".toCharArray();
    private TrustManagerFactory trustManagerFactory;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        StrictMode.ThreadPolicy policy = new StrictMode.ThreadPolicy.Builder().permitAll().build();
        StrictMode.setThreadPolicy(policy);
        setContentView(R.layout.activity_client_auth);
    }

    public static KeyPair generateKeyPair() {
        try {
            KeyPairGenerator keyGen = KeyPairGenerator.getInstance("RSA");
            keyGen.initialize(2048);
            return keyGen.genKeyPair();
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }

    protected static String pubKeyToPem(PublicKey publicKey) {
        byte[] keyBytes = publicKey.getEncoded();
        String header = "-----BEGIN PUBLIC KEY-----\n";
        String footer = "\n-----END PUBLIC KEY-----";
        String base64PublicKey = Base64.encodeToString(keyBytes, Base64.DEFAULT);
        return header + base64PublicKey + footer;
    }

    public KeyStore generateKeyStore(X509Certificate certificate, PrivateKey key) {
        try {
            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
            KeyStore keyStore = KeyStore.getInstance("PKCS12");
            keyStore.load(null, password);
            keyStore.setCertificateEntry("alias", certificate);
            keyStore.setKeyEntry("alias", key, password, new Certificate[]{certificate});
            keyStore.store(outputStream, password);
            outputStream.flush();
            KeyStore ks = KeyStore.getInstance("PKCS12");
            ks.load(new ByteArrayInputStream(outputStream.toByteArray()), password);
            return ks;
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }

    public static X509Certificate pemToCertificate(String pemCert) {
        try {
            InputStream inputStream = new ByteArrayInputStream(pemCert.getBytes(StandardCharsets.UTF_8));
            CertificateFactory certFactory = CertificateFactory.getInstance("X.509");
            return (X509Certificate) certFactory.generateCertificate(inputStream);
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }

    public static String inputStreamToString(InputStream stream, String charsetName) throws IOException {
        int n;
        char[] buffer = new char[1024 * 4];
        InputStreamReader reader = new InputStreamReader(stream, charsetName);
        StringWriter writer = new StringWriter();
        while (-1 != (n = reader.read(buffer))) writer.write(buffer, 0, n);
        return writer.toString();
    }

    public void nextRequest(View view) throws IOException {
        TextView textView = (TextView) findViewById(R.id.text);
        URL url;
        HttpsURLConnection urlConnection;
        int urlResponseCode;
        InputStream urlInputStream;
        String textViewText;
        String urlResponseText;

        HostnameVerifier hostnameVerifier = new HostnameVerifier() {
            @Override
            public boolean verify(String hostname, SSLSession session) {
                HostnameVerifier hv = HttpsURLConnection.getDefaultHostnameVerifier();
                return hv.verify("example.com", session);
            }
        };

        switch (requestNumber) {
            case 0: /* /gen-certificate/test */
                try {
                    KeyStore trustStore = KeyStore.getInstance("PKCS12");
                    trustStore.load(getAssets().open("truststore.p12"), "uselesspw".toCharArray());
                    trustManagerFactory = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
                    trustManagerFactory.init(trustStore);
                    context = SSLContext.getInstance("TLS");
                    context.init(null, trustManagerFactory.getTrustManagers(), null);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                KeyPair keyPair = generateKeyPair();
                assert keyPair != null;
                String pemPubKey = pubKeyToPem(keyPair.getPublic());

                url = new URL("https://192.168.2.2:4443/gen-certificate/test");
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setRequestMethod("POST");
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                urlConnection.setHostnameVerifier(hostnameVerifier);

                OutputStream urlOutputStream = urlConnection.getOutputStream();
                BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(urlOutputStream, "UTF-8"));
                writer.write("pubkey=" + URLEncoder.encode(pemPubKey, "utf-8"));
                writer.flush();
                writer.close();
                urlOutputStream.close();
                urlResponseCode = urlConnection.getResponseCode();
                urlInputStream = urlConnection.getInputStream();

                urlResponseText = inputStreamToString(urlInputStream, "UTF-8");
                generatedKeyStore = generateKeyStore(pemToCertificate(urlResponseText), keyPair.getPrivate());
                textViewText = "Request: https://service:4443/gen-certificate/test";
                textViewText = textViewText + "\nResponse status: " + urlResponseCode + "\nResponse text:\n" + urlResponseText;
                textViewText = textViewText + "\n\nNext request:\nhttps://service:4443/public (without client certificate)";
                textView.setText(textViewText);
                urlConnection.disconnect();
                break;
            case 1: /* /public (without client certificate) */
                try {
                    context = SSLContext.getInstance("TLS");
                    context.init(null, trustManagerFactory.getTrustManagers(), null);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                url = new URL("https://192.168.2.2:4443/public");
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                urlConnection.setHostnameVerifier(hostnameVerifier);

                urlResponseCode = urlConnection.getResponseCode();
                urlInputStream = urlConnection.getInputStream();

                urlResponseText = inputStreamToString(urlInputStream, "UTF-8");
                textViewText = "Request: https://service:4443/public (without client certificate)";
                textViewText = textViewText + "\nResponse status: " + urlResponseCode + "\nResponse text:\n" + urlResponseText;
                textViewText = textViewText + "\n\nNext request:\nhttps://service:4443/public (with previously generated certificate)";
                textView.setText(textViewText);
                urlConnection.disconnect();
                break;
            case 2: /* /public (with previously generated certificate) */
                try {
                    KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance("X509");
                    keyManagerFactory.init(generatedKeyStore, password);
                    context = SSLContext.getInstance("TLS");
                    context.init(keyManagerFactory.getKeyManagers(), trustManagerFactory.getTrustManagers(), null);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                url = new URL("https://192.168.2.2:4443/public");
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                urlConnection.setHostnameVerifier(hostnameVerifier);

                urlResponseCode = urlConnection.getResponseCode();
                urlInputStream = urlConnection.getInputStream();

                urlResponseText = inputStreamToString(urlInputStream, "UTF-8");
                textViewText = "Request: https://service:4443/public (with previously generated certificate)";
                textViewText = textViewText + "\nResponse status: " + urlResponseCode + "\nResponse text:\n" + urlResponseText;
                textViewText = textViewText + "\n\nNext request:\nhttps://service:4443/public (with correct certificate)";
                textView.setText(textViewText);
                urlConnection.disconnect();
                break;
            case 3: /* /public (with correct certificate) */
                try {
                    KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance("X509");
                    KeyStore keyStore = KeyStore.getInstance("PKCS12");
                    keyStore.load(getAssets().open("correct.p12"), password);
                    keyManagerFactory.init(keyStore, password);
                    context = SSLContext.getInstance("TLS");
                    context.init(keyManagerFactory.getKeyManagers(), trustManagerFactory.getTrustManagers(), null);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                url = new URL("https://192.168.2.2:4443/public");
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                urlConnection.setHostnameVerifier(hostnameVerifier);

                urlResponseCode = urlConnection.getResponseCode();
                urlInputStream = urlConnection.getInputStream();

                urlResponseText = inputStreamToString(urlInputStream, "UTF-8");
                textViewText = "Request: https://service:4443/public (with correct certificate)";
                textViewText = textViewText + "\nResponse status: " + urlResponseCode + "\nResponse text:\n" + urlResponseText;
                textViewText = textViewText + "\n\nNext request:\nhttps://service:4443/public (with revoked certificate)";
                textView.setText(textViewText);
                urlConnection.disconnect();
                break;
            case 4: /* /public (with revoked certificate) */
                try {
                    KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance("X509");
                    KeyStore keyStore = KeyStore.getInstance("PKCS12");
                    keyStore.load(getAssets().open("revoked.p12"), password);
                    keyManagerFactory.init(keyStore, password);
                    context = SSLContext.getInstance("TLS");
                    context.init(keyManagerFactory.getKeyManagers(), trustManagerFactory.getTrustManagers(), null);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                url = new URL("https://192.168.2.2:4443/public");
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                urlConnection.setHostnameVerifier(hostnameVerifier);

                urlResponseCode = urlConnection.getResponseCode();

                textViewText = "Request: https://service:4443/public (with revoked certificate)";
                textViewText = textViewText + "\nResponse status: " + urlResponseCode;
                textViewText = textViewText + "\n\nNext request:\nhttps://service:4443/public (with expired certificate)";
                textView.setText(textViewText);
                urlConnection.disconnect();
                break;
            case 5: /* /public (with expired certificate) */
                try {
                    KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance("X509");
                    KeyStore keyStore = KeyStore.getInstance("PKCS12");
                    keyStore.load(getAssets().open("expired.p12"), password);
                    keyManagerFactory.init(keyStore, password);
                    context = SSLContext.getInstance("TLS");
                    context.init(keyManagerFactory.getKeyManagers(), trustManagerFactory.getTrustManagers(), null);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                url = new URL("https://192.168.2.2:4443/public");
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                urlConnection.setHostnameVerifier(hostnameVerifier);
                try {
                    urlResponseCode = urlConnection.getResponseCode();
                    textView.setText(urlResponseCode);
                } catch (SSLHandshakeException e) {
                    textViewText = "Request: https://service:4443/public (with expired certificate)";
                    textViewText = textViewText + "\nHandshake failed (certificate is expired)";
                    textViewText = textViewText + "\n\nNext request:\nhttps://service:4443/public (with valid_but_not_accepted certificate)";
                    textView.setText(textViewText);
                }
                urlConnection.disconnect();
                break;
            case 6: /* /public (with valid_but_not_accepted certificate) */
                try {
                    KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance("X509");
                    KeyStore keyStore = KeyStore.getInstance("PKCS12");
                    keyStore.load(getAssets().open("valid_but_not_accepted.p12"), password);
                    keyManagerFactory.init(keyStore, password);
                    context = SSLContext.getInstance("TLS");
                    context.init(keyManagerFactory.getKeyManagers(), trustManagerFactory.getTrustManagers(), null);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                url = new URL("https://192.168.2.2:4443/public");
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                urlConnection.setHostnameVerifier(hostnameVerifier);
                try {
                    urlResponseCode = urlConnection.getResponseCode();
                    textView.setText(urlResponseCode);
                } catch (SSLHandshakeException e) {
                    textViewText = "Request: https://service:4443/public (with valid_but_not_accepted certificate)";
                    textViewText = textViewText + "\nHandshake failed (certificate not recognized)";
                    textViewText = textViewText + "\n\nNext request:\nhttps://service:4443/private (without client certificate)";
                    textView.setText(textViewText);
                }
                urlConnection.disconnect();
                break;
            case 7: /* /private (without client certificate) */
                try {
                    context = SSLContext.getInstance("TLS");
                    context.init(null, trustManagerFactory.getTrustManagers(), null);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                url = new URL("https://192.168.2.2:4443/private");
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                urlConnection.setHostnameVerifier(hostnameVerifier);

                urlResponseCode = urlConnection.getResponseCode();

                textViewText = "Request: https://service:4443/private (without client certificate)";
                textViewText = textViewText + "\nResponse status: " + urlResponseCode;
                textViewText = textViewText + "\n\nNext request:\nhttps://service:4443/private (with previously generated certificate)";
                textView.setText(textViewText);
                urlConnection.disconnect();
                break;
            case 8: /* /private  (with previously generated certificate) */
                try {
                    KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance("X509");
                    keyManagerFactory.init(generatedKeyStore, password);
                    context = SSLContext.getInstance("TLS");
                    context.init(keyManagerFactory.getKeyManagers(), trustManagerFactory.getTrustManagers(), null);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                url = new URL("https://192.168.2.2:4443/private");
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                urlConnection.setHostnameVerifier(hostnameVerifier);

                urlResponseCode = urlConnection.getResponseCode();
                urlInputStream = urlConnection.getInputStream();

                urlResponseText = inputStreamToString(urlInputStream, "UTF-8");
                textViewText = "Request: https://service:4443/private (with previously generated certificate)";
                textViewText = textViewText + "\nResponse status: " + urlResponseCode + "\nResponse text:\n" + urlResponseText;
                textViewText = textViewText + "\n\nNext request:\nhttps://service:4443/private (with correct certificate)";
                textView.setText(textViewText);
                urlConnection.disconnect();
                break;
            case 9: /* /private (with correct certificate) */
                try {
                    KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance("X509");
                    KeyStore keyStore = KeyStore.getInstance("PKCS12");
                    keyStore.load(getAssets().open("correct.p12"), password);
                    keyManagerFactory.init(keyStore, password);
                    context = SSLContext.getInstance("TLS");
                    context.init(keyManagerFactory.getKeyManagers(), trustManagerFactory.getTrustManagers(), null);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                url = new URL("https://192.168.2.2:4443/private");
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                urlConnection.setHostnameVerifier(hostnameVerifier);

                urlResponseCode = urlConnection.getResponseCode();
                urlInputStream = urlConnection.getInputStream();

                urlResponseText = inputStreamToString(urlInputStream, "UTF-8");
                textViewText = "Request: https://service:4443/private (with correct certificate)";
                textViewText = textViewText + "\nResponse status: " + urlResponseCode + "\nResponse text:\n" + urlResponseText;
                textViewText = textViewText + "\n\nNext request:\nhttps://service:4443/private (with revoked certificate)";
                textView.setText(textViewText);
                urlConnection.disconnect();
                break;
            case 10: /* /private (with revoked certificate) */
                try {
                    KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance("X509");
                    KeyStore keyStore = KeyStore.getInstance("PKCS12");
                    keyStore.load(getAssets().open("revoked.p12"), password);
                    keyManagerFactory.init(keyStore, password);
                    context = SSLContext.getInstance("TLS");
                    context.init(keyManagerFactory.getKeyManagers(), trustManagerFactory.getTrustManagers(), null);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                url = new URL("https://192.168.2.2:4443/private");
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                urlConnection.setHostnameVerifier(hostnameVerifier);

                urlResponseCode = urlConnection.getResponseCode();

                textViewText = "Request: https://service:4443/private (with revoked certificate)";
                textViewText = textViewText + "\nResponse status: " + urlResponseCode;
                textViewText = textViewText + "\n\nNext request:\nhttps://service:4443/public (with expired certificate)";
                textView.setText(textViewText);
                urlConnection.disconnect();
                break;
            case 11: /* /private (with expired certificate) */
                try {
                    KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance("X509");
                    KeyStore keyStore = KeyStore.getInstance("PKCS12");
                    keyStore.load(getAssets().open("expired.p12"), password);
                    keyManagerFactory.init(keyStore, password);
                    context = SSLContext.getInstance("TLS");
                    context.init(keyManagerFactory.getKeyManagers(), trustManagerFactory.getTrustManagers(), null);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                url = new URL("https://192.168.2.2:4443/private");
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                urlConnection.setHostnameVerifier(hostnameVerifier);
                try {
                    urlResponseCode = urlConnection.getResponseCode();
                    textView.setText(urlResponseCode);
                } catch (SSLHandshakeException e) {
                    textViewText = "Request: https://service:4443/private (with expired certificate)";
                    textViewText = textViewText + "\nHandshake failed (certificate is expired)";
                    textViewText = textViewText + "\n\nNext request:\nhttps://service:4443/private (with valid_but_not_accepted certificate)";
                    textView.setText(textViewText);
                }
                urlConnection.disconnect();
                break;
            case 12: /* /private (with valid_but_not_accepted certificate) */
                try {
                    KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance("X509");
                    KeyStore keyStore = KeyStore.getInstance("PKCS12");
                    keyStore.load(getAssets().open("valid_but_not_accepted.p12"), password);
                    keyManagerFactory.init(keyStore, password);
                    context = SSLContext.getInstance("TLS");
                    context.init(keyManagerFactory.getKeyManagers(), trustManagerFactory.getTrustManagers(), null);
                } catch (Exception e) {
                    e.printStackTrace();
                }

                url = new URL("https://192.168.2.2:4443/private");
                urlConnection = (HttpsURLConnection) url.openConnection();
                urlConnection.setSSLSocketFactory(context.getSocketFactory());
                urlConnection.setHostnameVerifier(hostnameVerifier);
                try {
                    urlResponseCode = urlConnection.getResponseCode();
                    textView.setText(urlResponseCode);
                } catch (SSLHandshakeException e) {
                    textViewText = "Request: https://service:4443/private (with valid_but_not_accepted certificate)";
                    textViewText = textViewText + "\nHandshake failed (certificate not recognized)";
                    textView.setText(textViewText);
                }
                urlConnection.disconnect();
                break;
            case 13:
                textViewText = "There are no more requests to be made.";
                textView.setText(textViewText);
                break;
        }
        requestNumber++;
    }
}