package br.com.nubank.jetty;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Scanner;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;

import static java.util.concurrent.TimeUnit.HOURS;

public class CRLUpdater {
    protected static final String s3Crl = "crl/crl.txt";
    private static final ScheduledExecutorService updater = Executors.newScheduledThreadPool(1);

    protected static void updateCRL(S3FileRetriever s3FileRetriever) throws ParseException {
        ConcurrentHashMap<String, Date> newCrl = new ConcurrentHashMap<>();
        SimpleDateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ");
        Scanner sc = new Scanner(s3FileRetriever.getS3FileInputStream(s3Crl));
        while (sc.hasNextLine()) {
            String line = sc.nextLine();
            String[] pair = line.split(" ");
            Date date = df.parse(pair[1].replaceAll("Z", "+0000"));
            newCrl.put(pair[0], date);
        }
        CertificateRevocationHandler.setCrl(newCrl);
    }

    protected static void startUpdateCrlScheduler() {
        final Runnable updateCrlRunnable = () -> {
            try {
                updateCRL();
            } catch (Exception e) {
                e.printStackTrace();
            }
        };
        updater.scheduleAtFixedRate(updateCrlRunnable, 1, 1, HOURS);
    }

    protected static void updateCRL() throws ParseException {
        updateCRL(new S3FileRetriever());
    }
}
